#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>
#include <time.h>
#include <syslog.h>
#include <string.h>
#include <dlfcn.h>

extern char *optarg;

#include <sndfile.h>

#include "listener.h"
#include "error.h"
#include "snd_dev.h"
#include "utils.h"
#include "lib.h"

char *wav_path = NULL; 					/* path for wav-files */
char *dev_name = NULL; 					/* path to audio-device */
char *configfile = NULL;				/* configurationfile */

int detect_level = 754;					/* threshold for sound-detection */
int rec_silence = 2;					/* number of seconds to keep on reading when silence is detected */
int min_duration = 1;					/* how many samples above threshold before recording starts */
int max_duration = -1;					/* max. duration IN SECONDS(!) */
int sample_rate = SAMPLE_RATE;				/* speaks for itself */
int compression = SF_FORMAT_PCM_16;			/* default is no compression */
int format = SF_FORMAT_WAV;				/* default is WAV output */
char channels = 1;					/* mono or stereo */
char *exec = NULL;					/* what to execute after recording */

int n_filter_lib = 0;					/* filters */
filter_lib filter[MAX_N_LIBRARIES];
char safe_after_filter = 0;

char do_exit = 0;

int get_compression_type( const char *pcCompression )
{
	if (strcasecmp(pcCompression, "u-law") == 0)
		return SF_FORMAT_ULAW;
	if (strcasecmp(pcCompression, "a-law") == 0)
		return SF_FORMAT_ALAW;
	if (strcasecmp(pcCompression, "IMA-ADPCM") == 0)
		return SF_FORMAT_IMA_ADPCM;
	if (strcasecmp(pcCompression, "MS-ADPCM") == 0)
		return SF_FORMAT_MS_ADPCM;
	if (strcasecmp(pcCompression, "GSM-6.10") == 0)
		return SF_FORMAT_GSM610;
	if (strcasecmp(pcCompression, "G721_32") == 0)
		return SF_FORMAT_G721_32;
	if (strcasecmp(pcCompression, "G723_24") == 0)
		return SF_FORMAT_G723_24;
	if (strcasecmp(pcCompression, "G723_40") == 0)
		return SF_FORMAT_G723_40;

	return -1;
} /* int get_compression_type( ) */

int get_format_type(const char *type)
{
	if (strcasecmp(type, "wav") == 0)
		return SF_FORMAT_WAV;
	if (strcasecmp(type, "aiff") == 0)
		return SF_FORMAT_AIFF;
	if (strcasecmp(type, "au") == 0)
		return SF_FORMAT_AU;
	if (strcasecmp(type, "raw") == 0)
		return SF_FORMAT_RAW;
	if (strcasecmp(type, "svx") == 0)
		return SF_FORMAT_SVX;
	if (strcasecmp(type, "nist") == 0)
		return SF_FORMAT_NIST;
	if (strcasecmp(type, "voc") == 0)
		return SF_FORMAT_VOC;
	if (strcasecmp(type, "ircam") == 0)
		return SF_FORMAT_IRCAM;
	if (strcasecmp(type, "w64") == 0)
		return SF_FORMAT_W64;
	if (strcasecmp(type, "mat4") == 0)
		return SF_FORMAT_MAT4;
	if (strcasecmp(type, "mat5") == 0)
		return SF_FORMAT_MAT5;
	if (strcasecmp(type, "pvf") == 0)
		return SF_FORMAT_PVF;
	if (strcasecmp(type, "xi") == 0)
		return SF_FORMAT_XI;
	if (strcasecmp(type, "htk") == 0)
		return SF_FORMAT_HTK;
	if (strcasecmp(type, "sds") == 0)
		return SF_FORMAT_SDS;
	if (strcasecmp(type, "avr") == 0)
		return SF_FORMAT_AVR;
	if (strcasecmp(type, "wavex") == 0)
		return SF_FORMAT_WAVEX;

	return -1;
}

void start_wavfile(SNDFILE **sndfile_out, char *fname)
{
	SF_INFO sfinfo_out;
	struct tm *tmdummy;
	time_t now;
	char *ts, *dummy;

	/* create filename */
	time(&now);
	tmdummy = localtime(&now);
	sprintf(fname, "%s/%04d-%02d-%02d_%02d%02d%02d.wav",
		wav_path,
		tmdummy -> tm_year + 1900,
		tmdummy -> tm_mon + 1,
		tmdummy -> tm_mday,
		tmdummy -> tm_hour,
		tmdummy -> tm_min,
		tmdummy -> tm_sec);

	/* create file */
	memset(&sfinfo_out, 0x00, sizeof(sfinfo_out));
        sfinfo_out.samplerate = sample_rate;
        sfinfo_out.channels = channels;
        sfinfo_out.format = format | compression;
        *sndfile_out = sf_open(fname, SFM_WRITE, &sfinfo_out);
	if (!*sndfile_out)
		error_exit("Could not create file %s!", fname);

	/* add string describing 'listener' to file */
	(void)sf_set_string(*sndfile_out, SF_STR_SOFTWARE, "Generated with listener v" VERSION ", http://www.vanheusden.com/listener/");
	/* add timestamp to file */
	ts = ctime(&now);
	dummy = strchr(ts, '\n');
	if (dummy) *dummy = 0x00;
	(void)sf_set_string(*sndfile_out, SF_STR_DATE, ts);

	syslog(LOG_INFO, "Started recording to %s", fname);
}

void start_wav_file_processor(char *fname)
{
	if (exec)
	{
		pid_t pid;

		syslog(LOG_INFO, "Starting childprocess: %s", exec);

		pid = fork();

		if (pid == -1)
		{
			syslog(LOG_ERR, "Failed to fork! %m");
		}
		else if (pid == 0)
		{
			if (-1 == execlp(exec, exec, fname, (void *)NULL))
				error_exit("Failed to start childprocess %s with parameter %s: %m", exec, fname);
		}
	}
}

void record(int fd, short *first, int first_n_samples)
{
	SNDFILE *sndfile_out = NULL;
	char fname[PATH_MAX];
	time_t recstart, prev_sound_start, prev_max_duration, now;
	short *buffer;

	buffer = (short *)mymalloc(sample_rate * sizeof(short) * channels, "recording buffer(2)");

	start_wavfile(&sndfile_out, fname);

	time(&recstart);
	prev_max_duration = prev_sound_start = now = recstart;

	/* write sample to disk */
	if (sf_writef_short(sndfile_out, first, first_n_samples) != first_n_samples)
		error_exit("failed to write to wav-file\n");

	for(; (now - prev_sound_start) < rec_silence && do_exit == 0; )
	{
		int loop, threshold_count = 0;

		/* read sample */
		if (READ(fd, (char *)buffer, sample_rate * sizeof(short) * channels) != (sample_rate * sizeof(short) * channels))
			error_exit("failed to read from audio-device\n");

		if (!safe_after_filter)
		{
			/* write sample to disk */
			if (sf_writef_short(sndfile_out, buffer, sample_rate * channels) != sample_rate)
				error_exit("failed to write to wav-file\n");
		}

		/* do filters (if any) */
		/* do filtering AFTER writing sample to disk */
		for(loop=0; loop<n_filter_lib; loop++)
		{
			filter[loop].do_filter(filter[loop].lib_data, (void *)buffer, sample_rate);
		}

		if (safe_after_filter)
		{
			/* write sample to disk */
			if (sf_writef_short(sndfile_out, buffer, sample_rate * channels) != sample_rate)
				error_exit("failed to write to wav-file\n");
		}

		/* see if it became silent again */
		for(loop=0; loop<(sample_rate * channels); loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					time(&prev_sound_start);
					break;
				}
			}
		}

		time(&now);

		/* took too long? */
		if (max_duration != -1 && (now - prev_max_duration) > max_duration)
		{
			sf_close(sndfile_out);

			start_wav_file_processor(fname);

			start_wavfile(&sndfile_out, fname);

			prev_max_duration = now;
		}
	}

	if (sndfile_out)
	{
		sf_close(sndfile_out);

		start_wav_file_processor(fname);
	}

	free(buffer);

	syslog(LOG_INFO, "Stopped recording (%d seconds)", (int)(now - recstart));
}

void sigh(int sig)
{
	if (sig == SIGTERM)
		do_exit = 1;
}


void prt_usage( const char *pcProgram )
{
	printf( "\n"
		"Usage: %s [options]\n"
		"-C<compression>  Set WAV compression    -e<command>      Script to call after recording\n"
		"-r<rate>         Sample rate            -m<min_duration> Mini. duration to record (samples)\n"
		"-R<rec_silence>  Silence to record      -l<detect_level> Detect level\n"
		"-c<configfile>   Configfile to use      -x<max_duration> Max. duration to record (seconds)\n"
		"-w<wave-dir>     Where to write .WAVs   -d<device>       DSP device to use (/dev/dsp)\n"
		"-z<channels>     Number of channels (1 (default) or 2)\n"
		"-t<format>       Output format (see manual)\n"
		"-p               Read from pipe (together with splitaudio)\n"
		"-s               Be silent              -h               This help text\n\n",
			pcProgram );
}

/* This specifies the minimum user-supplied settings we need without a 
 * specified configuration-file.
 */
#define MAX_SETTINGS	6
int main(int argc, char *argv[])
{
	int	c;
	int	silent = 0;	/* be silent on stdout? */
	char	show_help = 0;
	char	from_pipe = 0;
	char	set_from_pipe = 0, set_sample_rate = 0, set_rec_silence = 0, set_detect_level = 0, set_max_duration = 0, 
		set_min_duration = 0, set_compression = 0, set_exec = 0, set_dev_name = 0, set_wav_path = 0, set_channels = 0,
		set_format = 0;
	int	fd;
	int	loop;
	FILE	*fh;
	/* right now I'm not sure if a short is always 2 bytes. check this on your
	 * platform
	 */
	short	*buffer;
	int	buffer_size;

	/* Check command-line options;
	 */
	while ( (c = getopt( argc, argv, "c:shS::r:R:l:m:C:e:w:d:x:t:" )) != -1 )
	{
		switch (c)
		{
		  case 'w':	/* WAV directory */
			wav_path = optarg;
			set_wav_path = 1;
		 	break;

		  case 'd':	/* Device */
			dev_name = optarg;
			set_dev_name = 1;
			break;

		  case 'e':	/* Exec */
			exec = optarg;
			set_exec = 1;
			break;

		  case 'C':	/* Compression */
			if ((compression = get_compression_type(optarg)) == -1)
			{
				fprintf(stderr, "%s is not a known compression type\n", optarg);
				return 2;
			}
			set_compression = 1;
			break;

                  case 't':	/* format type */
			if ((format = get_format_type(optarg)) == -1)
			{
				fprintf(stderr, "%s is an unknown output format\n", optarg);
				return 2;
			}
			set_format = 1;
			break;

		  case 'm':	/* Min duration */
			min_duration = atoi(optarg);
			set_min_duration = 1;
			break;

		  case 'x':	/* Max duration */
			max_duration = atoi(optarg);
			set_max_duration = 1;
			break;

		  case 'l':	/* Detect level */
			detect_level = atoi(optarg);
			set_detect_level = 1;
			break;

		  case 'R':	/* Record silence */
			rec_silence = atoi(optarg);
			set_rec_silence = 1;
			break;

		  case 'r':	/* Sample rate */
			sample_rate = atoi(optarg);
			set_sample_rate = 1;
			break;
			
		  case 'h':	/* Help -- display usage */
			show_help = 1;
			break;

		  case 'c':	/* Configuration-file */
			configfile = strdup(optarg);
			break;

		  case 's':	/* Be silent.. */
			silent = 1;
			break;

		  case 'p':	/* read from pipe */
			from_pipe = 1;
			set_from_pipe = 1;
			break;

		  case 'z':	/* number of channels */
			channels = atoi(optarg);
			set_channels = 1;
			break;

		  default:
			/* unknown switch */
			show_help = 1;
			break;
		} /* switch.. */	

	} /* while.. */

	if (silent == 0)
		printf("listener v" VERSION ", (C)2003-2004 by folkert@vanheusden.com\n");

	if (from_pipe && channels > 1)
	{
		fprintf(stderr, "You can only monitor 1 channel when reading from a pipe!\n");
		return 1;
	}

	if (show_help == 1)
	{
		prt_usage(argv[0]);	
		return 1;
	}
	
	/* Apply default settings;
	 */
	if (configfile == NULL) configfile = strdup( CONFIGFILE );
	if (dev_name == NULL) dev_name = strdup( DSP_PATH );
	if (wav_path == NULL) wav_path = strdup( WAV_PATH );

	fh = fopen(configfile, "rb");
	if (!fh) 
		error_exit("error opening configfile %s\n", configfile);

	for(;;)
	{
		char *i_is, *cmd, *par = NULL;
		char *line = read_line(fh);
		if (!line)
			break;

                /* find '=' */
                i_is = strchr(line, '=');
                if (i_is)
                {
                        /* find start of parameter(s) */
                        par = i_is+1;
                        while ((*par) == ' ') par++;
                }
                cmd = get_token(line);

		if (strcasecmp(cmd, "wav_path") == 0)
		{
			if (!set_wav_path)
				wav_path = strdup(par);
		}
		else if (strcasecmp(cmd, "devname") == 0)
		{
			if (!set_dev_name)
				dev_name = strdup(par);
		}
		else if (strcasecmp(cmd, "detect_level") == 0)
		{
			if (!set_detect_level)
				detect_level = atoi(par);
		}
		else if (strcasecmp(cmd, "rec_silence") == 0)
		{
			if (!set_rec_silence)
				rec_silence = atoi(par);
		}
		else if (strcasecmp(cmd, "min_duration") == 0)
		{
			if (!set_min_duration)
				min_duration = atoi(par);
		}
		else if (strcasecmp(cmd, "max_duration") == 0)
		{
			if (!set_max_duration)
				max_duration = atoi(par);
		}
		else if (strcasecmp(cmd, "sample_rate") == 0)
		{
			if (!set_sample_rate)
				sample_rate = atoi(par);
		}
		else if (strcasecmp(cmd, "from_pipe") == 0)
		{
			if (!set_from_pipe)
				from_pipe = atoi(optarg);
		}
		else if (strcasecmp(cmd, "compression") == 0)
		{
			if (!set_compression && (compression = get_compression_type( par )) == -1)
				error_exit("%s is an unknown compression format\n");
		}
		else if (strcasecmp(cmd, "format") == 0)
		{
			if (!set_format && (format = get_format_type( par )) == -1)
				error_exit("%s is an unknown output format\n");
		}
		else if (strcasecmp(cmd, "exec") == 0)
		{
			if (!set_exec)
				exec = strdup(par);
		}
		else if (strcasecmp(cmd, "channels") == 0)
		{
			if (!set_channels)
				channels = atoi(par);
		}
		else if (strcasecmp(cmd, "filter") == 0)
		{
			if (n_filter_lib < MAX_N_LIBRARIES)
			{
                                char *dummy = strchr(par, ' ');

                                dlerror();

                                if (dummy) *dummy = 0x00;
				filter[n_filter_lib].library = dlopen(par, RTLD_NOW);
				if (!filter[n_filter_lib].library)
					error_exit("Failed to load filter library %s, reason: %s\n", par, dlerror());

				filter[n_filter_lib].init_library = dlsym(filter[n_filter_lib].library, "init_library");
				if (!filter[n_filter_lib].init_library)
					error_exit("The filter library %s is missing the 'init_library' function");

				filter[n_filter_lib].do_filter    = dlsym(filter[n_filter_lib].library, "do_filter");
				if (!filter[n_filter_lib].do_filter)
					error_exit("The filter library %s is missing the 'do_library' function");

                                filter[n_filter_lib].par = dummy+1 ? strdup(dummy+1) : NULL;

				n_filter_lib++;
			}
			else
				error_exit("Too many filters defined, only %d possible.\n", MAX_N_LIBRARIES);
		}
		else if (strcasecmp(cmd, "safe_after_filter") == 0)
		{
			if (strcasecmp(par, "yes") == 0 ||
			    strcasecmp(par, "on") == 0 ||
			    strcasecmp(par, "1") == 0)
			{
				safe_after_filter = 1;
			}
			else
			{
				safe_after_filter = 0;
			}
		}
		else if (cmd[0] != '#')
			error_exit("'%s=%s' is not a known configuration-statement\n", cmd?cmd:"(none)", par?par:"(none)");

		free(cmd);
		free(line);
	}
	fclose(fh);

	if (min_duration < 1)
		error_exit("min_duration must be at least 1!");

	/* open audio-device */
	fd = open(dev_name, O_RDWR);
	if (fd == -1)
		error_exit("Error opening device %s\n", dev_name);

	/* set audio-device to 44.1KHz 16bit mono */
	if (!from_pipe)
	{
		set_soundcard_sample_format(fd);
		set_soundcard_number_of_channels(fd, channels);
		set_soundcard_sample_rate(fd, sample_rate);
	}

	if (silent == 0)
	{
		printf("Path:         %s\n", wav_path);
		printf("Device:       %s\n", dev_name);
		printf("Level:        %d\n", detect_level);
		printf("Rec silence:  %d\n", rec_silence);
		printf("Min duration: %d\n", min_duration);
		printf("Samplerate:   %d\n", sample_rate);
		printf("Channels:     %d\n", channels);

		if (from_pipe)
			printf("Reading from pipe\n");
	}

	/* initialize filters */
	for(loop=0; loop<n_filter_lib; loop++)
	{
		filter[loop].lib_data = filter[loop].init_library(sizeof(short), channels, sample_rate, filter[loop].par);
	}

	if (daemon(-1, -1) == -1)
		error_exit("problem becoming daemon process\n");

	syslog(LOG_INFO, "listener started");

	buffer_size = sample_rate * sizeof(short) * sizeof(channels);
	buffer = (short *)mymalloc(buffer_size, "recording buffer(1)");

	signal(SIGTERM, sigh);
	signal(SIGCHLD, SIG_IGN);

	for(;do_exit == 0;)
	{
		int threshold_count = 0;
		int rc, loop;

		/* check audio-level */
		/* read data */
		rc = READ(fd, (char *)buffer, buffer_size);
		if (rc < 0)
			error_exit("error while reading from audio-device!\n", rc);

		/* do filters (if any) */
		for(loop=0; loop<n_filter_lib; loop++)
		{
			filter[loop].do_filter(filter[loop].lib_data, (void *)buffer, sample_rate);
		}

		/* check audio-levels */
		for(loop=0; loop<(sample_rate * channels); loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					record(fd, buffer, sample_rate);
					break;
				}
			}
		}
	}

	return 0;
}
